@@ -46,23 +46,38 @@ BEGIN {
 
 sub fret
 {
-    my ($package) = @_;
+    my ($package);
+    ($package, @ARGV) = @_;
     my ($font, $maxx, $maxy, $pdf, $root);
     my (%opt);
     my ($fh, $fdat);
+    
+    if ($ARGV[0] =~ /^-p(.*)$/oi)
+    {
+    	$package = $1;
+    	shift @ARGV;
+    	$package = shift @ARGV unless $package;
+    }
+
+	$package = 'Font::Fret::Default' unless $package;
+	
+	if ($package->can('process_argv'))
+	{	$package->process_argv (\%opt); }
+	else
+	{   getopts("d:fgh:m:p:qrs:", \%opt); }
 
-    getopts("fgh:m:p:qrs:", \%opt);
 
     unless (defined $ARGV[0])
     {
-        die <<'EOT';
-FRET [-f] [-g] [-r] [-s size] [-p package] [-q] font_file [out_file]
+    	die <<'EOT';
+fret [-p package] [-f] [-g] [-r] [-s size] [-q] font_file [out_file]
 Generates a report on a font according to a particular package. In some
 contexts the package may be over-ridden. Paper size may also be specified.
 
 If no out_file is given then out_file becomes font_file.pdf (removing .ttf
 if present)
 
+  -d num        Sets time for testing purposes
   -f            Don't try to save memory on large fonts (>1000 glyphs)
   -g            Add one glyph per page report following summary report
   -h            Mode for glyph per page output. Bitfield:
@@ -73,13 +88,16 @@ if present)
   -q            quiet mode
   -r            Don't output report lines, fill the page with glyph boxes
   -s size       paper size: a4, ltr, legal
+
+If supplied, -p must be the first option.
 EOT
+						# ' quote matching for above here-doc.
+
     }
 
     $opt{s} = lc($opt{s}) || 'ltr';
     $opt{s} = 'ltr' unless defined $sizes{$opt{s}};
     ($maxx, $maxy) = @{$sizes{$opt{s}}};
-    $package = $opt{p} || $package || 'Font::Fret::Default';
 
     unless (defined $ARGV[1])
     {
@@ -208,7 +226,7 @@ no strict;
     $ftrleft = "BT 1 0 0 1 36 27 Tm 80 Tz /FR 7 Tf (FRET v$VERSION "
         . "Package $package " . ${"${package}::VERSION"} . ") Tj ET\n";
 use strict;
-    @time = split(/\s+/, localtime());
+    @time = split(/\s+/, localtime($opt{d} || time()));
     $tr = "Printed at $time[3] on $time[0] $time[2] $time[1] $time[4]   Page ";
     @time = split(/\s+/, localtime($font->{'head'}->getdate));
     $hdrrw = "Modified at $time[3] on $time[0] $time[2] $time[1] $time[4]";
@@ -280,7 +298,7 @@ use strict;
                    . ($ybase + 65.5) . " l s [] 0 d\n");
             }
             $yorg = $ybase + 32 - ($font->{'head'}{'yMax'} + $font->{'head'}{'yMin'}) * $tsize / $upem / 2;
-            $ppage->add("$dots 36 $yorg m " . (36 + $numperow * 54) . " $yorg l S [] 0 d\n") if ($yorg > $ybase);
+            $ppage->add(sprintf("%s 36 %.4f m %d %.4f l S [] 0 d\n", $dots, $yorg, 36 + $numperow * 54, $yorg)) if ($yorg > $ybase);
             $xcentre = 63;
             for ($i = 0; $i < $numperow; $i++, $xcentre += 54)
             {
@@ -297,22 +315,16 @@ use strict;
                     $xorg = ($glyph->{'xMax'} + $glyph->{'xMin'}) * $tsize / $upem;
                     $xorg = $xcentre - $xorg / 2;
                     $xadv = $xorg + $font->{'hmtx'}{'advance'}[$gid] * $tsize / $upem;
-                    $ppage->add("$dots $xadv " . ($ybase + 7) . " m $xadv "
-                            . ($ybase + 57) . " l S [] 0 d\n") if ($xadv < $xcentre + 27);
-                    $ppage->add("$dots $xorg " . ($ybase + 7) . " m $xorg "
-                            . ($ybase + 57) . " l S [] 0 d\n") if ($xorg > $xcentre - 27
-                            && $xorg < $xcentre + 27);
-                    $ppage->add("BT 1 0 0 1 $xorg $yorg Tm /T$id $tsize Tf 100 Tz " .
-                            $gcol . sprintf("<%04X> Tj " .
-                            ($gcol ? "0 g " : "") . "ET\n", $gid));
+                    $ppage->add(sprintf("%s %.4f %d m %.4f %d l S [] 0 d\n", $dots, $xadv, $ybase + 7, $xadv,$ybase + 57)) if ($xadv < $xcentre + 27);
+                    $ppage->add(sprintf("%s %.4f %d m %.4f %d l S [] 0 d\n", $dots, $xorg, $ybase + 7, $xorg, $ybase + 57)) if ($xorg > $xcentre - 27 && $xorg < $xcentre + 27);
+                    $ppage->add(sprintf("BT 1 0 0 1 %.4f %.4f Tm /T$id $tsize Tf 100 Tz $gcol <%04X> Tj %s ET\n", $xorg, $yorg, $gid, ($gcol ? "0 g " : "")));
                     unless ($opt{'r'})
                     {
                         $gxorg = ($glyph->{'xMax'} + $glyph->{'xMin'}) * $gsize / $upem / 2;
                         $gxorg = 274 - $gxorg;
                         $gyorg = $ybase + (3 - $i) * 16 + 8 - ($font->{'head'}{'yMax'} +
                                 $font->{'head'}{'yMin'}) * $gsize / $upem / 2;;
-                        $ppage->add("BT 1 0 0 1 $gxorg $gyorg Tm /T$id $gsize Tf 100 Tz " .
-                                sprintf("<%04X> Tj ET\n", $gid));
+                        $ppage->add(sprintf("BT 1 0 0 1 %.4f %.4f Tm /T$id $gsize Tf 100 Tz <%04X> Tj ET\n", $gxorg, $gyorg, $gid));
                     }
                 }
                 @parms = ($cids[$gcount + $i], $gid, $glyph, $rev[$gid], $font);
@@ -462,7 +474,7 @@ use strict;
                 { $ppage->add(sprintf("%.2f %.2f m %s\n", $x, $y, $blob)); }
             } else
             { $ppage->add(sprintf("%.2f %.2f m %s\n", $x, $y, $offblob)); }
-            if ($txt ne '' && $opt{'h'} & 1 == 0)
+            if ($txt ne '' && ($opt{'h'} & 1) == 0)
             {
                 $tw = $pdf_helv->width($txt) * 4.8 + 2;         # 6pt + 2pt margin
                 $tx = $x + ($dirs[$j][0] > 0 ? 0 : -$tw);
@@ -476,10 +488,23 @@ use strict;
                         $tx, $ty, asPDFStr($txt)));
             }
         }
+        if (($opt{'h'} & 2) == 0)
+        {
+            my ($points) = $package->extra_points($font, $gid, $glyph);
 
+            foreach my $p (@{$points})
+            {
+                $x = ($p->[0] - $fxmin) * $optgsize / $upem + 58;
+                $y = ($p->[1] - $fymin) * $optgsize / $upem + 144;
+                $ppage->add(sprintf("%.2f %.2f m %s\n", $x, $y, $bigblob));
+                $ppage->add(sprintf("BT 1 0 0 1 %.2f %.2f Tm 80 Tz /FR 8 Tf %s Tj ET\n",
+                        $x + 3, $y - 3, asPDFStr($p->[2])));
+            }
+        }
         $ppage->{' curstrm'}{'Filter'} = PDFArray(PDFName('FlateDecode'));
         $ppage->ship_out($pdf);
         $ppage->empty;
+        $glyph->empty;
     }
 }
 
@@ -635,12 +660,12 @@ sub out_row
                 if $info[2] =~ m/([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})/oi;
         $f = "BT $col 1 0 0 1 ";
         if ($info[0] =~ m/r/oi)
-        { $f .= "$xr"; }
+        { $f .= sprintf("%.4f", $xr); }
         elsif ($info[0] =~ m/c/oi)
-        { $f .= "$xc"; }
+        { $f .= sprintf("%.4f", $xc); }
         else
-        { $f .= "$xl"; }
-        $f .= " $yorg Tm ";
+        { $f .= sprintf("%.4f", $xl); }
+        $f .= sprintf(" %.4f Tm ", $yorg);
         $g = PDFStr($e);
         $f .= "/F$ft $pt Tf 80 Tz " . $g->as_pdf . " Tj ET";
         $f .= " 0 g" if ($col);
@@ -967,3 +992,32 @@ sub label
     return '';
 }
 
+
+=head2 extra_points
+
+Returns an array of arrays for a list of extra points to display on the glyph page.
+Each sub array consists of [x, y, label]
+
+=cut
+
+sub extra_points
+{
+    my ($class, $font, $gid, $glyph) = @_;
+
+    return [];
+}
+
+=head1 AUTHOR
+
+Martin Hosken L<Martin_Hosken@sil.org>. 
+
+=head1 LICENSING
+
+Copyright (c) 1998-2013, SIL International (http://www.sil.org)
+
+This module is released under the terms of the Artistic License 2.0.
+For details, see the full text of the license in the file LICENSE.
+
+=cut
+
+1;
