@@ -24,7 +24,6 @@ Reference to a L<font|Font::TTF::Font> structure. C<read_font> will cause at lea
 the L<post|Font::TTF::Post>, L<cmap|Font::TTF::Cmap>, L<loca|Font::TTF::Loca>, and 
 L<name|Font::TTF::Name> tables to be read in.
 
-
 =item glyphs
 
 An array of references to glyph data structures, indexed by glyphID. Stucture elements are:
@@ -33,7 +32,7 @@ An array of references to glyph data structures, indexed by glyphID. Stucture el
 
 =item uni
 
-Unicode scalar value, if any, as specified in the APDB. (decimal integer)
+Array of Unicode scalar values (decimal integers), if any, that map from cmap to this glyph. 
 
 =item gnum
 
@@ -43,11 +42,17 @@ Actual glyph ID from font.
 
 Actual Postscript name from font.
 
+=item name
+
+This element is set by L</"make_names"> or L</"make_classes"> and is the replacement name returned 
+by L</"make_name">.
+
 =back
 
 Note: The C<uni>, C<gnum> and C<post> values are based on the C<UID>, C<GID>, and C<PSName> fields
 of the APDB. If there are descrepancies between the APDB and the font's internal tables, then 
-for calcuating the above three values, priority is given first to C<UID> field, then C<PSName> field, and finally C<GID>. 
+for calcuating the above three values, priority is given first to C<UID> field, then C<PSName> field, 
+and finally C<GID>. 
 
 =over 4
 
@@ -96,6 +101,10 @@ comma separated list of bounding box coordinates, i.e., C<x1, y1, x2, y2>
 
 Unicode scalar value, if any, of the component. (decimal integer)
 
+=item line
+
+Line number in APDB where this component is defined.
+
 =back
 
 =back
@@ -146,7 +155,7 @@ each bit set to 1 if the corresponding glyph has the given attachment point.
 
 =item ligclasses
 
-Optionally created by make_classes if ligatures are requested and they exist. The base forms class is no_I<code> while the ligatures are held in I<code>.
+Optionally created by L</"make_classes"> if ligatures are requested and they exist. The base forms class is no_I<code> while the ligatures are held in I<code>.
 
 =item WARNINGS
 
@@ -154,7 +163,7 @@ If C<-errorfh> not set, this accumulates any warning or error messages encounter
 
 =item cWARNINGS
 
-Count of number fo warnings or errors encountered.
+Count of number of warnings or errors encountered.
 
 =back
 
@@ -162,13 +171,18 @@ Count of number fo warnings or errors encountered.
 
 =cut
 
+# ' make editors happy
+
 use Font::TTF::Font 0.36;
 use XML::Parser::Expat;
+use Algorithm::Diff qw(sdiff);
 
 use strict;
 use vars qw($VERSION);
 
-$VERSION = "0.07";  # MH    add make_names if you don't want make_classes
+$VERSION = "0.09";  # MH  Add classes property support
+# $VERSION = "0.08";  # BH	Generalize values for %opts so can be space- or comma-separated list in a scalar, or can be an array ref.
+# $VERSION = "0.07";  # MH    add make_names if you don't want make_classes
 # $VERSION = "0.06";  # MH    debug glyph alternates for ligature creation, add Unicode
 # $VERSION = "0.05";  # MH    add glyph alternates e.g. A/u0410 and ligature class creation
 # $VERSION = "0.04";	# BH   in progress
@@ -199,7 +213,8 @@ Options that may be supplied throught the C<%opts> hash include:
 
 =item -omittedAPs
 
-A comma-separated list of attachment point types to ignore.
+A list of attachment point types to ignore. Can be a string containing comma- or space-separated names,
+or a ref to an array of strings. 
 
 =item -strictap
 
@@ -208,8 +223,10 @@ points on the outline of the glyph.
 
 =item -knownemptyglyphs
 
-A comma-separated list of names of glyphs that are known to have no outline 
-(thus shouldn't generate warning).
+If this option is specified, C<read_font> will warn if glyphs that should have outlines don't.
+The option value should be a list of names of glyphs that are known to have no outline 
+(thus shouldn't generate warning). Can be a string containing comma- or space-separated names,
+or a ref to an array of strings.
 
 =item -errorfh
 
@@ -223,91 +240,145 @@ warning messages are accumulated in C<WARNINGS>.
 sub read_font
 {
     my ($class, $fname, $xml_file, %opts) = @_;
-    my (@glyphs, $f, $t, $xml, $cur_glyph, $cur_pt);
+    my (@glyphs, $f, $t, $xml, $cur_glyph, $cur_pt, $numg, @reverse);
     my ($self) = {};
     bless $self, ref $class || $class;
 
     my (%omittedAPs, %known_empty_glyphs);
-    map {$omittedAPs{$_} = $omittedAPs{"_$_"} = 1} split (',', $opts{'-omittedAPs'});
-    map {$known_empty_glyphs{$_} = 1} split (',', $opts{'-knownemptyglyphs'});
+    map {$omittedAPs{$_} = 1} ( ref ($opts{'-omittedAPs'}) eq 'ARRAY' ? @{$opts{'-omittedAPs'}} : ($opts{'-omittedAPs'} =~m/[^\s,]+/go));
+    map {$omittedAPs{"_$_"} = 1} grep {/^[^_]/} keys %omittedAPs;
+    map {$known_empty_glyphs{$_} = 1} ( ref ($opts{'-knownemptyglyphs'}) eq 'ARRAY' ? @{$opts{'-knownemptyglyphs'}} : ($opts{'-knownemptyglyphs'} =~m/[^\s,]+/go));
 
-    $f = Font::TTF::Font->open($fname) || die "Can't open font $fname";
-    foreach $t (qw(post cmap loca name))
-    { $f->{$t}->read; }
+    if ($fname)
+    {
+        $f = Font::TTF::Font->open($fname) || die "Can't open font $fname";
+        foreach $t (qw(post cmap loca name))
+        { $f->{$t}->read; }
 
-    $self->{'font'} = $f;
-    $self->{'cmap'} = $f->{'cmap'}->find_ms->{'val'} || die "Can't find Unicode table in font $fname";
-    my (@reverse) = $f->{'cmap'}->reverse('array' => 1);
+        $self->{'font'} = $f;
+        $self->{'cmap'} = $f->{'cmap'}->find_ms->{'val'} || die "Can't find Unicode table in font $fname";
+        @reverse = $f->{'cmap'}->reverse('array' => 1);
+        $numg = $f->{'maxp'}{'numGlyphs'};
+    }
+
+    
+#    my $minUID;
+#    if (exists $f->{'OS/2'})
+#    {
+#    		my $os2 = $f->{'OS/2'}->read || die "Can't read OS/2 table in font $fname";
+#    		$minUID = $os2->{'usFirstCharIndex'};
+#    }
+#	printf STDERR "FirstCharIndex = U+%04X\n", $minUID;
 
     $xml = XML::Parser::Expat->new();
     $xml->setHandlers('Start' => sub {
         my ($xml, $tag, %attrs) = @_;
 
-        if ($tag eq 'glyph')
+        if ($tag eq 'font')
+        {
+            foreach (qw(name upem))
+            { $self->{$_} = $attrs{$_}; }
+        }
+        elsif ($tag eq 'glyph')
         {
             my ($ug, $pg, $ig);
             $cur_glyph = {%attrs};
             undef $cur_pt;
 
-            if (defined $attrs{'UID'})
+            if ($f && defined $attrs{'UID'})
             {
+                $attrs{'UID'} =~ s/^U\+//o;      # Not supposed to contain "U+", but some do
                 my ($uni) = hex($attrs{'UID'});
                 $ug = $self->{'cmap'}{$uni};
-                $self->error($xml, $cur_glyph, undef, "No glyph associated with UID $attrs{'UID'}") unless (defined $ug);
-                $cur_glyph->{'gnum'} = $ug;
+                if (defined $ug)
+                { $cur_glyph->{'gnum'} = $ug; }
+                else
+                { $self->APerror($xml, $cur_glyph, undef, "No glyph associated with UID $attrs{'UID'}"); }
                 $cur_glyph->{'uni'} = [$uni];
                 # delete $attrs{'UID'};  # Added in MH's version; v0.04: now believed un-needed and un-wanted.
             }
-            if (defined $attrs{'PSName'})
+            if ($f && defined $attrs{'PSName'})
             {
                 $pg = $f->{'post'}{'STRINGS'}{$attrs{'PSName'}};
-                $self->error($xml, $cur_glyph, undef, "No glyph associated with postscript name $attrs{'PSName'}") unless (defined $pg);
-                $self->error($xml, $cur_glyph, undef, "Postscript name: $attrs{'PSName'} resolves to different glyph to Unicode ID: $attrs{'UID'}")
-                        if (defined $attrs{'UID'} && $pg != $ug);
-                $cur_glyph->{'gnum'} ||= $pg;
+                unless (defined $pg)
+                {
+                	# Failed to find glyph by the supplied PSName -- see if this is one of two special cases.
+                	# These cases exist because FontLab doesn't use the correct (Apple-specified) name for U+000D and U+00A7
+                	$pg = $f->{'post'}{'STRINGS'}{'nonmarkingreturn'} if $attrs{'PSName'} eq 'CR';
+                	$pg = $f->{'post'}{'STRINGS'}{'macron'} 			if $attrs{'PSName'} eq 'overscore';
+                }
+				if (defined $pg)
+				{
+                	$self->APerror($xml, $cur_glyph, undef, "Postscript name: $attrs{'PSName'} resolves to different glyph to Unicode ID: $attrs{'UID'}")
+	                        if (defined $ug && $pg != $ug);
+	                $cur_glyph->{'gnum'} ||= $pg;
+	            }
+                else
+                {
+                	$self->APerror($xml, $cur_glyph, undef, "No glyph associated with postscript name $attrs{'PSName'}") ;
+                }
                 # delete $attrs{'PSName'};  # Added in MH's version; v0.04: now believed un-needed and un-wanted.
             }
             if (defined $attrs{'GID'})
             {
                 $ig = $attrs{'GID'};
-                $self->error($xml, $cur_glyph, undef, "Specified glyph id $attrs{'GID'} different to glyph of Unicode ID: $attrs{'UID'}")
-                        if (defined $attrs{'UID'} && $ug != $ig);
-                $self->error($xml, $cur_glyph, undef, "Specified glyph id $attrs{'GID'} different to glyph of postscript name $attrs{'PSName'}")
-                        if (defined $attrs{'PSName'} && $pg != $ig);
+                $self->APerror($xml, $cur_glyph, undef, "Specified glyph id $attrs{'GID'} different to glyph of Unicode ID: $attrs{'UID'}")
+                        if (defined $ug && $ug != $ig);
+                $self->APerror($xml, $cur_glyph, undef, "Specified glyph id $attrs{'GID'} different to glyph of postscript name $attrs{'PSName'}")
+                        if (defined $pg && $pg != $ig);
+                $self->APerror($xml, $cur_glyph, undef, "Specified glyph id $attrs{'GID'} is >= number of glyphs in font ($numg)")
+                        if ($ig < 0 || ($f && $ig >= $numg));
                 $cur_glyph->{'gnum'} ||= $ig;
                 # delete $attrs{'GID'}; # Added in MH's version; v0.04: now believed un-needed and un-wanted.
             }
-            $cur_glyph->{'post'} = $f->{'post'}{'VAL'}[$cur_glyph->{'gnum'}];
-            $cur_glyph->{'uni'} = $reverse[$cur_glyph->{'gnum'}] if (!defined $cur_glyph->{'uni'} && defined $reverse[$cur_glyph->{'gnum'}]);
-            $cur_glyph->{'PSName'} = $cur_glyph->{'post'} if ($cur_glyph->{'post'} && $cur_glyph->{'post'} ne '.notdef');
-
-            if ($cur_glyph->{'glyph'} = $f->{'loca'}{'glyphs'}[$cur_glyph->{'gnum'}])
+            if ($f)
             {
-                # v0.04: Slight difference in this code and MH's: this code causes
-                # $cur_glyph->{'glyph'} to be defined for all glyphs; in MH's code
-                # it was defined only for non-empty glyphs.
-                $cur_glyph->{'glyph'}->read_dat;
-                if ($cur_glyph->{'glyph'}{'numberOfContours'} > 0)
-                { $cur_glyph->{'props'}{'drawn'} = 1; }
-                $cur_glyph->{'glyph'}->get_points;
+                $cur_glyph->{'post'} = $f->{'post'}{'VAL'}[$cur_glyph->{'gnum'}];
+                $cur_glyph->{'uni'} = $reverse[$cur_glyph->{'gnum'}] if (!defined $cur_glyph->{'uni'} && defined $reverse[$cur_glyph->{'gnum'}]);
+
+                if ($cur_glyph->{'glyph'} = $f->{'loca'}{'glyphs'}[$cur_glyph->{'gnum'}])
+                {
+                    # v0.04: Slight difference in this code and MH's: this code causes
+                    # $cur_glyph->{'glyph'} to be defined for all glyphs; in MH's code
+                    # it was defined only for non-empty glyphs.
+                    $cur_glyph->{'glyph'}->read_dat;
+                    if ($cur_glyph->{'glyph'}{'numberOfContours'} > 0)
+                    { $cur_glyph->{'props'}{'drawn'} = 1; }
+                    $cur_glyph->{'glyph'}->get_points;
+                }
+                elsif ($opts{'-knownemptyglyphs'})
+                {
+                    $self->APerror($xml, $cur_glyph, undef, "Empty glyph outline in font") unless $known_empty_glyphs{$cur_glyph->{'post'}};
+                }
+
+                # MH's code includes the following two lines, but these are redundant with 
+                # assignment $cur_glyph = {%attrs} at start of this block
+                #foreach (keys %attrs)
+                #{ $cur_glyph->{$_} = $attrs{$_}; }
             }
             else
             {
-                $self->error($xml, $cur_glyph, undef, "No glyph outline in font") unless $known_empty_glyphs{$cur_glyph->{'post'}};
+                my (@basicnames) = qw(.notdef .null cr);
+#                $numg ||= 3;
+                if ($cur_glyph->{'gnum'} > $numg)
+                { $numg = $cur_glyph->{'gnum'} + 1; }
+                else
+                { $cur_glyph->{'gnum'} ||= $numg++; }
+
+                $cur_glyph->{'post'} = $cur_glyph->{'PSName'} || $basicnames[$cur_glyph->{'gnum'}];
+                $cur_glyph->{'uni'} = [map {hex($_)} split(' ', $cur_glyph->{'UID'})] if (defined $cur_glyph->{'UID'});
             }
-
-            # MH's code includes the following two lines, but these are redundant with 
-            # assignment $cur_glyph = {%attrs} at start of this block
-            #foreach (keys %attrs)
-            #{ $cur_glyph->{$_} = $attrs{$_}; }
-
             $cur_glyph->{'line'} = $xml->current_line;
             $self->{'glyphs'}[$cur_glyph->{'gnum'}] = $cur_glyph;
 
         } elsif ($tag eq 'compound')
         {
-            my $component = {%attrs};
-            $component->{'uni'} = [hex($attrs{'UID'})] if defined $attrs{'UID'};
+            my $component = {%attrs, line => $xml->current_line};
+            if (defined $attrs{'UID'})
+            {
+                $attrs{'UID'} =~ s/^U\+//o;      # Not supposed to contain "U+", but some do
+                $component->{'uni'} = [hex($attrs{'UID'})] ;
+            }
             push @{$cur_glyph->{'components'}}, $component;
         } elsif ($tag eq 'point')
         {
@@ -323,12 +394,12 @@ sub read_font
             my ($cont) = $attrs{'num'};
             my ($g) = $cur_glyph->{'glyph'} || return;
 
-            $self->error($xml, $cur_glyph, $cur_pt, "Specified contour of $cont different from calculated contour of $cur_pt->{'cont'}")
+            $self->APerror($xml, $cur_glyph, $cur_pt, "Specified contour of $cont different from calculated contour of $cur_pt->{'cont'}")
                     if (defined $cur_pt->{'cont'} && $cur_pt->{'cont'} != $attrs{'num'});
 
             if (($cont == 0 && $g->{'endPoints'}[0] != 0)
                 || ($cont > 0 && $g->{'endPoints'}[$cont-1] + 1 != $g->{'endPoints'}[$cont]))
-            { $self->error($xml, $cur_glyph, $cur_pt, "Contour $cont not a single point path"); }
+            { $self->APerror($xml, $cur_glyph, $cur_pt, "Contour $cont not a single point path"); }
             else
             { $cur_pt->{'cont'} = $cont; }
 
@@ -341,7 +412,7 @@ sub read_font
             my ($g) = $cur_glyph->{'glyph'};
             my ($cont, $i);
 
-            $self->error($xml, $cur_glyph, $cur_pt, "Specified location of ($x, $y) different from calculated location ($cur_pt->{'x'}, $cur_pt->{'y'})")
+            $self->APerror($xml, $cur_glyph, $cur_pt, "Specified location of ($x, $y) different from calculated location ($cur_pt->{'x'}, $cur_pt->{'y'})")
                     if (defined $cur_pt->{'x'} && ($cur_pt->{'x'} != $x || $cur_pt->{'y'} != $y));
 
             if ($g)
@@ -357,15 +428,15 @@ sub read_font
                     }
                 }
                 if ($g->{'x'}[$i] != $x || $g->{'y'}[$i] != $y)
-                { $self->error($xml, $cur_glyph, $cur_pt, "No glyph point at specified location ($x, $y)") if ($opts{'-strictap'}); }
+                { $self->APerror($xml, $cur_glyph, $cur_pt, "No glyph point at specified location ($x, $y)") if ($opts{'-strictap'}); }
                 if (($cont == 0 && $g->{'endPoints'}[0] != 0)
                     || $g->{'endPoints'}[$cont-1] + 1 != $g->{'endPoints'}[$cont])
-                { $self->error($xml, $cur_glyph, $cur_pt, "Calculated contour $cont not a single point path") if ($opts{'-strictap'}); }
+                { $self->APerror($xml, $cur_glyph, $cur_pt, "Calculated contour $cont not a single point path") if ($opts{'-strictap'}); }
                 else
                 { $cur_pt->{'cont'} = $cont; }
             }
             else
-            { $self->error($xml, $cur_glyph, $cur_pt, "No glyph point at specified location ($x, $y)") if ($opts{'-strictap'}); }
+            { $self->APerror($xml, $cur_glyph, $cur_pt, "No glyph point at specified location ($x, $y)") if ($opts{'-strictap'}); }
 
             $cur_pt->{'x'} = $x unless defined $cur_pt->{'x'};
             $cur_pt->{'y'} = $y unless defined $cur_pt->{'y'};
@@ -385,8 +456,8 @@ sub read_font
         undef $xml;
     }
 
+    $self->{'numg'} = $numg;
 # now fill in the glyphs that aren't in the xml
-    my ($numg) = $f->{'maxp'}{'numGlyphs'};
     my ($i);
 
     for ($i = 0; $i < $numg; $i++)
@@ -394,10 +465,11 @@ sub read_font
         next if (defined $self->{'glyphs'}[$i]);
 
         my ($cur_glyph) = {'gnum' => $i};
+        $self->{'glyphs'}[$i] = $cur_glyph;
+        next unless ($f);
+
         $cur_glyph->{'uni'} = $reverse[$i] if (defined $reverse[$i]);
         $cur_glyph->{'post'} = $f->{'post'}{'VAL'}[$i];
-        $cur_glyph->{'PSName'} = $cur_glyph->{'post'} if ($cur_glyph->{'post'} && $cur_glyph->{'post'} ne '.notdef');
-        $self->{'glyphs'}[$i] = $cur_glyph;
         if ($cur_glyph->{'glyph'} = $f->{'loca'}{'glyphs'}[$i])
         {
             # v0.04: Slight difference in this code and MH's: this code causes
@@ -408,9 +480,9 @@ sub read_font
             { $cur_glyph->{'props'}{'drawn'} = 1; }
             $cur_glyph->{'glyph'}->get_points;
         }
-        else
+        elsif ($opts{'-knownemptyglyphs'})
         {
-            $self->error($xml, $cur_glyph, undef, "No glyph outline in font") unless $known_empty_glyphs{$cur_glyph->{'post'}};
+            $self->APerror($xml, $cur_glyph, undef, "Empty glyph outline in font") unless $known_empty_glyphs{$cur_glyph->{'post'}};
         }
     }
     $self;
@@ -418,15 +490,16 @@ sub read_font
 
 =head2 $ap->make_names
 
-Create name records for all the glyphs in the font
+An alternative to L</"make_classes">, this method just creates name records for all the glyphs in the font. 
+That is, for every glyph record in C<glyphs>, L</"make_names"> invokes L</"make_name"> and saves the result 
+in the C<name> element of the glyph.
 
 =cut
 
 sub make_names
 {
     my ($self) = @_;
-    my ($f) = $self->{'font'};
-    my ($numg) = $f->{'maxp'}{'numGlyphs'};
+    my ($numg) = $self->{'numg'};
     my ($i, $gname);
 
     for ($i = 0; $i < $numg; $i++)
@@ -436,7 +509,7 @@ sub make_names
         $gname = $self->make_name($glyph->{'post'}, $glyph->{'uni'}, $glyph);
 
         while (defined $self->{'glyph_names'}{$gname})
-        { $gname =~ s/(?:_(\d+))$/"_".($1+1)/oe; }
+        { $gname =~ s/(?:_(\d+))?$/"_".($1+1)/oe; }
         $self->{'glyph_names'}{$gname} = $i;
         $glyph->{'name'} = $gname;
     }
@@ -459,10 +532,14 @@ Options supported are:
 
 =item -ligatures
 
-Takes two values: first or last. First creates ligature classes with the class based on the first element of the ligature and the contents of the class on the rest of the ligature. Last creates classes based on the last element of the ligature, thus grouping all glyphs with the same last ligature element together. Ligature classes are stored in C<$self->{'ligclasses'}>.
+Accepts one of two possible values: C<'first'> or C<'last'>. First creates ligature classes with the class based on the first element of the ligature and the contents of the class on the rest of the ligature. Last creates classes based on the last element of the ligature, thus grouping all glyphs with the same last ligature element together. Ligature classes are stored in C<$self->{'ligclasses'}>.
 
 Ligature elements are separated by _ in the glyph name. Ligatures are only made if there are corresponding non ligature glyphs in the font. A final .text on the glyph name of a ligature is assumed to be associated with the whole ligature and not just the last element.
 
+=item -ligtype
+
+If set to C<'comp'>, then treat glyph name extension as part of the final element.
+
 =back
 
 =cut
@@ -470,11 +547,11 @@ Ligature elements are separated by _ in the glyph name. Ligatures are only made
 sub make_classes
 {
     my ($self, %opts) = @_;
-    my ($f) = $self->{'font'};
+    my ($numg) = $self->{'numg'};
     my (%classes, %namemap);
     my ($g, $gname, $i, $j, $glyph, %used, $p, $name);
 
-    for ($i = 0; $i < $f->{'maxp'}{'numGlyphs'}; $i++)
+    for ($i = 0; $i < $numg; $i++)
     {
         $glyph = $self->{'glyphs'}[$i];
         $gname = $self->make_name($glyph->{'post'}, $glyph->{'uni'}, $glyph);
@@ -489,7 +566,7 @@ sub make_classes
 
         foreach $p (keys %{$glyph->{'points'}})
         {
-            my ($pname) = $self->make_point($p, $glyph);
+            my ($pname) = $self->make_point($p, $glyph, \%opts);
             next unless ($pname);                           # allow for point deletion, in effect.
             if ($p ne $pname)
             {
@@ -498,9 +575,20 @@ sub make_classes
             }
             push (@{$self->{'lists'}{$pname}}, $i);
             vec($self->{'vecs'}{$pname}, $i, 1) = 1 if ($self->{'vecs'});
+            next if (defined $opts{'-notmark'} and $opts{'-notmark'} =~ m/\b$pname\b/);
+            vec($self->{'ismarks'}, $i, 1) = 1 if ($pname =~ m/^_/o);
         }
         foreach (split('/', $glyph->{'post'}))
         { $namemap{$_} = $i; }
+        if (defined $glyph->{'props'}{'classes'})
+        {
+            my ($c);
+            foreach $c (split(' ', $glyph->{'props'}{'classes'}))
+            {
+                $c =~ s/^c//o;
+                push (@{$classes{$c}}, $glyph->{'gnum'});
+            }
+        }
     }
 
     # need a separate loop since using other glyphs' names
@@ -512,8 +600,9 @@ sub make_classes
                                             # when processing x.y, etc.
             {
                 my ($base, $ext) = ($` , $1);    #` make editor happy
-                if ($i = $namemap{$base})
+                if ($base && (!defined $classes{"no_$ext"} || defined $namemap{$base}))
                 {
+                    my ($i) = $namemap{$base};
                     push (@{$classes{$ext}}, $glyph->{'gnum'});
                     push (@{$classes{"no_$ext"}}, $self->{'glyphs'}[$i]{'gnum'});
                 }
@@ -537,18 +626,17 @@ sub make_classes
                 if ($opts{'-ligatures'} eq 'first')
                 { 
                     $class = $elem[0];
-                    $base = "uni$base" if ($class =~ s/^uni//o);
-                    $base =~ s/^_//o;
+                    $base =~ s/^_u?//o;
                 }
                 else
                 { 
                     $class = $elem[-1];
-                    $class =~ s/^_//o;
+                    $class =~ s/^_u?//o;
                 }
 
+                next unless ($i = $namemap{$base});
                 $cname = $class;
                 $cname =~ s/\./_/og;
-                next unless ($i = $namemap{$base});
                 unless (defined $self->{'ligmap'}{$cname})
                 {
                     my ($match) = 0;
@@ -580,11 +668,14 @@ could be overridden when subclassing.
 
 =cut
 
+#  ' make editors happy
+
 sub make_name
 {
     my ($self, $gname, $uni, $glyph) = @_;
     $gname =~ s{/.*$}{}o;           # strip alternates
-    $gname = sprintf("u%04x", $uni) if ($gname eq '.notdef');
+    $gname = defined $uni ? sprintf("u%04x", $uni->[0]) : "glyph$glyph->{'gnum'}" if $gname eq '.notdef';
+    $gname ||= ".unk";
     $gname;
 }
 
@@ -599,10 +690,127 @@ By default this returns $pname, but the function could be overridden when subcla
 
 sub make_point
 {
-    my ($self, $p, $glyph) = @_;
+    my ($self, $p, $glyph, $opts) = @_;
+    return undef if ($opts->{'-ignoredAPs'} and $opts->{'-ignoredAPs'} =~ m/\b$p\b/);
     $p;
 }
 
+=head2 $fv->align_glyphs($dat)
+
+Provides one possible way to map glyph names in a new font with an existing project.
+
+Compares the list of glyphs found during L</"read_font"> (the I<new> glyph list) with the list 
+from L</"parse_volt"> in $dat (the I<old> glyph list>, returning an array that provides
+a mapping between old and new glyph IDs. The array is indexed by old glyph ID and returns the
+new glyph ID.
+
+This implementation works by C<sdiff>ing the two ordered lists of glyph C<name>s and 
+then matching them up, first by C<name> then by C<uni>.
+
+=cut
+
+sub align_glyphs
+{
+    my ($self, $data) = @_;
+    my (@map, @revmap, @old, @new, @diff, $s, $g, $u);
+
+    @new = map {$_ ? $_->{'name'} : ""} @{$self->{'glyphs'}};
+    @old = map {$_ ? $_->{'name'} : ""} @{$data->{'glyphs'}};
+    @diff = sdiff(\@old, \@new);
+
+# first find solid alignments
+    foreach $s (@diff)
+    {
+        if ($s->[0] eq 'u')
+        {
+            $map[$data->{'glyph_names'}{$s->[1]}] = $self->{'glyph_names'}{$s->[2]};
+            $revmap[$self->{'glyph_names'}{$s->[2]}] = $data->{'glyph_names'}{$s->[1]};
+        }
+    }
+
+# now deal with the rest
+    foreach $s (@diff)
+    {
+        if ($s->[0] eq '-' || $s->[0] eq 'c')
+        {
+            my ($gnum) = $data->{'glyph_names'}{$s->[1]};
+            my ($uni) = $data->{'glyphs'}[$gnum]{'uni'};
+            my ($gnew);
+    # anything with the same name not already used?
+            if ($g = $self->{'glyph_names'}{$s->[1]} and !defined $revmap[$g])
+            {
+                $map[$gnum] = $g;
+                $revmap[$g] = $gnum;
+            }
+    # anything spare with the same unicode (any unicode the same)?
+            else
+            {
+                foreach $g (@{$self->{'glyphs'}})
+                {
+                    if ($g->{'uni'} == $uni && !defined $revmap[$g->{'gnum'}])
+                    {
+                        $gnew = $g->{'gnum'};
+                        last;
+                    }
+                }
+                if ($gnew)
+                {
+                    $map[$gnum] = $gnew;
+                    $revmap[$gnew] = $gnum;
+                    last;
+                }
+            }
+    # make it a deletion (i.e. in old but not in new)
+        }
+        elsif ($s->[0] eq '+')
+        {
+            # nothing to do since only really interested in old->new mapping
+        }
+    }
+
+# deal with unaligned conflicts and simply align them
+    foreach $s (@diff)
+    {
+        if ($s->[0] eq 'c')
+        {
+            my ($gnum) = $data->{'glyph_names'}{$s->[1]};
+            my ($nnum) = $self->{'glyph_names'}{$s->[2]};
+            next if ($map[$gnum] || $revmap[$nnum]);
+            $map[$gnum] = $nnum;
+            $revmap[$nnum] = $gnum;
+        }
+    }
+    return (@map);
+}
+
+=head2 $str = $ap->as_xml()
+
+Returns an XML representation of the AP database
+
+=cut
+
+sub as_xml
+{
+    my ($self) = @_;
+    my ($numg) = $self->{'numg'};
+    my ($res, $i);
+
+    $res = "<?xml version='1.0'?>\n<font";
+    foreach (qw(name upem))
+    { $res .= " $_='$self->{$_}'" if (defined $self->{$_}); }
+    $res .= ">\n\n";
+
+    for ($i = 0; $i < $numg; $i++)
+    {
+        my ($g) = $self->{'glyphs'}[$i];
+        $res .= "<glyph GID='$i'";
+        $res .= " PSName='$g->{'post'}'" if (defined $g->{'post'});
+        $res .= " UID='" . join(' ', map{sprintf("%04X", $_)} @{$g->{'uni'}}) . "'" if (defined $g->{'uni'});
+        $res .= ">\n";
+    }
+
+    $res;
+}
 # Private routine:'
 
 sub split_lig
@@ -620,25 +828,25 @@ sub split_lig
         { $_ = "_$_"; }
         $base = $str;
         if ($type =~ /last/)
-        { $base =~ s/_(.*?)$//o; }
+        { $base =~ s/_([^_]*)$//o; }
         else
         { $base =~ s/^(.*?)_//o; }
     }
     elsif ($str =~ s/^uni//o)
     {
         @res = $str =~ m/([0-9a-fA-F]{4})/og;
-        $res[0] = "uni$res[0]";
         if ($type =~ /last/)
-        { $base = "uni" . join('', @res[0 .. ($#res-2)]); }
+        { $base = "uni" . join('', @res[0 .. ($#res-1)]); }
         else
-        { $base = "uni" . join('', @res[1 .. ($#res-1)]); }
+        { $base = "uni" . join('', @res[1 .. $#res]); }
+        $res[0] = "uni$res[0]";
     }
     else
     { $res[0] = $str; }
     ($ext, $base, @res);
 }
 
-sub error
+sub APerror
 {
     my $self = shift;
     my ($xml, $cur_glyph, $cur_pt, $str) = @_;
@@ -661,7 +869,14 @@ sub error
 
     $msg .=  " at line " . $xml->current_line if ($xml);
     $msg .= ".\n";
+    $self->error($msg);
+}
 
+
+sub error
+{
+    my $self = shift;
+    my $msg = join(' ', @_);
     if (defined $self->{'-errorfh'})
     { print {$self->{'-errorfh'}} $msg; }
     else
@@ -670,8 +885,22 @@ sub error
     $self->{'cWARNINGS'}++;
 }
 
+
+1;
+
 =head1 See also
 
 L<TTFBuilder|bin::TTFBuilder>, L<Font::TTF::Font>
 
-=cut
+=head1 AUTHOR
+
+Martin Hosken L<Martin_Hosken@sil.org>. 
+
+=head1 LICENSING
+
+Copyright (c) 1998-2013, SIL International (http://www.sil.org)
+
+This module is released under the terms of the Artistic License 2.0.
+For details, see the full text of the license in the file LICENSE.
+
+=cut
\ No newline at end of file
