@@ -36,8 +36,8 @@ sub out_gdl
     {
         $glyph = $self->{'glyphs'}[$i];
         $fh->print("$glyph->{'name'} = ");
-        if ($opts{'-psnames'} && $glyph->{'PSName'})
-        { $fh->print("postscript(\"$glyph->{'PSName'}\")"); }
+        if ($opts{'-psnames'} && $glyph->{'post'} && $glyph->{'post'} ne '.notdef')
+        { $fh->print("postscript(\"$glyph->{'post'}\")"); }
         else
         { $fh->print("glyphid($i)"); }
 
@@ -122,7 +122,7 @@ sub out_gdl
 
 sub out_classes
 {
-    my ($self, $fh) = @_;
+    my ($self, $fh, %opts) = @_;
     my ($f) = $self->{'font'};
     my ($lists) = $self->{'lists'};
     my ($classes) = $self->{'classes'};
@@ -142,6 +142,7 @@ sub out_classes
         else
         { $name =~ s/^_//o; }
 
+        $fh->print("#define HAS_c${name}Dia 1\n") if ($opts{'-defines'} && $name !~ m/^Takes/o);
         $fh->print("c${name}Dia = (");
         $count = 0; $sep = '';
         foreach $cl (@{$lists->{$l}})
@@ -165,6 +166,7 @@ sub out_classes
             next if ($psname eq '' || $psname eq '.notdef');
             next if (vec($vecs->{$l}, $c, 1));
             next if (defined $glyphs->[$c]{'props'}{'GDL_order'} && $glyphs->[$c]{'props'}{'GDL_order'} <= 1);
+            next unless (vec($self->{'ismarks'}, $c, 1));
             $fh->print("$sep$glyphs->[$c]{'name'}");
             if (++$count % 8 == 0)
             { $sep = ",\n    "; }
@@ -177,6 +179,7 @@ sub out_classes
 
     foreach $cl (sort {classcmp($a, $b)} keys %{$classes})
     {
+        $fh->print("#define HAS_c$cl 1\n") if ($opts{'-defines'} && $cl !~ m/^no_/o);
         $fh->print("c$cl = ($glyphs->[$classes->{$cl}[0]]{'name'}");
         for ($i = 1; $i <= $#{$classes->{$cl}}; $i++)
         { $fh->print($i % 8 ? ", $glyphs->[$classes->{$cl}[$i]]{'name'}" : ",\n    $glyphs->[$classes->{$cl}[$i]]{'name'}"); }
@@ -185,6 +188,7 @@ sub out_classes
 
     foreach $cl (sort {classcmp($a, $b)} keys %{$ligclasses})
     {
+        $fh->print("#define HAS_clig$cl 1\n") if ($opts{'-defines'} && $cl !~ m/^no_/o);
         $fh->print("clig$cl = ($glyphs->[$ligclasses->{$cl}[0]]{'name'}");
         for ($i = 1; $i <= $#{$ligclasses->{$cl}}; $i++)
         { $fh->print($i % 8 ? ", $glyphs->[$ligclasses->{$cl}[$i]]{'name'}" : ",\n    $glyphs->[$ligclasses->{$cl}[$i]]{'name'}"); }
@@ -213,15 +217,24 @@ sub endtable
 
 sub end_gdl
 {
-    my ($self, $fh, $include) = @_;
+    my ($self, $fh, $include, $defines) = @_;
+    my ($k, $v);
 
+    while (($k, $v) = each %$defines)
+    {
+        $fh->print("\n#define $k $v");
+    }
     $fh->print("\n#define MAXGLYPH " . ($self->{'font'}{'maxp'}{'numGlyphs'} - 1) . "\n");
-    $fh->print("\n#include \"$include\"\n") if ($include);
+    foreach (@$include)
+    {
+        $fh->print("\n#include \"$_\"\n");
+    }
 }
 
 sub make_name
 {
     my ($self, $gname, $uni, $glyph) = @_;
+    $gname =~ s/[:\(\)\{\}]//g;
     $gname =~ s{/.*$}{}o;
     $gname =~ s/\.(.)/'_'.lc($1)/oge;
     if ($gname =~ m/^u(?:[0-9A-Fa-f]{4,6})/oi)
@@ -234,7 +247,7 @@ sub make_name
     {
         my (@nums) = $gname =~ m/([0-9A-Fa-f]{4})/og;
         $gname =~ s/[0-9A-Fa-f]{4}//og;
-        $gname = 'g_' . join('_', @nums) . $gname;
+        $gname = 'g' . join('_', map {lc($_)} @nums) . $gname;
     }
     else
     {
@@ -246,7 +259,7 @@ sub make_name
 
 sub make_point
 {
-    my ($self, $p, $glyph) = @_;
+    my ($self, $p, $glyph, $opts) = @_;
 
     if ($p =~ m/^%([a-z0-9]+)_([a-z0-9]+)$/oi)
     {
@@ -261,6 +274,7 @@ sub make_point
         return undef;
     }
 
+    return undef if ($opts->{'-ignoredAPs'} and $opts->{'-ignoredAPs'} =~ m/\b$p\b/);
     return $p;
 }
 
@@ -309,25 +323,76 @@ sub lig_rules
 {
     my ($self, $fh, $pnum, $type) = @_;
     my ($ligclasses) = $self->{'ligclasses'};
-    my ($c);
+    my ($c, %namemap, $glyph, $name);
 
     return unless (defined $pnum);
-    return unless (scalar %{$self->{'ligclasses'}});
     $fh->print("\ntable(substitution);\npass($pnum);\n");
-    foreach $c (grep {!m/^no_/o} keys %{$ligclasses})
+    if (scalar %$ligclasses)
     {
-        my ($gnum) = $self->{'ligmap'}{$c};
-        my ($gname) = $self->{'glyphs'}[$gnum]{'name'};
-        my ($compstr);
+        foreach $c (grep {!m/^no_/o} keys %{$ligclasses})
+        {
+            my ($gnum) = $self->{'ligmap'}{$c};
+            my ($gname) = $self->{'glyphs'}[$gnum]{'name'};
+            my ($compstr);
 
-        if ($self->{'glyphs'}[$ligclasses->{$c}[0]]{'compounds'}{'0'})
-        { $compstr = ' {component.0.reference = @1; component.1.reference = @2}'; }
+            if ($self->{'glyphs'}[$ligclasses->{$c}[0]]{'compounds'}{'0'})
+            { $compstr = ' {component.0.reference = @1; component.1.reference = @2}'; }
 
-        if ($type eq 'first')
-        { $fh->print("$gname clno_$c > _ cl$c:(1 2)$compstr / _ ^ _;\n"); }
-        else
-        { $fh->print("clno_$c $gname > cl$c:(1 2)$compstr _/ ^ _ _;\n"); }
+            if ($type eq 'first')
+            { $fh->print("$gname cligno_$c > _ clig$c:(1 2)$compstr / _ ^ _;\n"); }
+            else
+            { $fh->print("cligno_$c $gname > clig$c:(1 2)$compstr _ / ^ _ _;\n"); }
 
+        }
+    }
+
+    foreach $glyph (@{$self->{'glyphs'}})
+    {
+        foreach $name (split('/', $glyph->{'post'}))
+        { $namemap{$name} = $glyph; }
+    }
+    foreach $glyph (@{$self->{'glyphs'}})
+    {
+        foreach $name (split('/', $glyph->{'post'}))
+        {
+            my ($ext, $base, @elem) = $self->split_lig($name, $type, '');
+            next if ($ext || scalar @elem < 3);
+            my ($islig) = 1;
+            my (@parts, $e, $g);
+            my ($class, $oglyph);
+            if ($type eq 'first')
+            { $class = $elem[0]; }
+            else
+            {
+                $class = $elem[-1];
+                $class =~ s/^_//g;
+            }
+            $class =~ s/\./_/g;
+            $oglyph = $namemap{$base};
+
+            foreach $e (@elem)
+            {
+                my ($n) = $e;
+                my ($g);
+                $n =~ s/_//g;
+                foreach ($n, "uni$n", "u$n")
+                {
+                    if ($g = $namemap{$_})
+                    { last; }
+                }
+                if (!$g)
+                {
+                    $islig = 0;
+                    last;
+                }
+                else
+                {
+                    push(@parts, $g->{'name'});
+                }
+            }
+            next if (!$islig);
+            $fh->print(join(" ", @parts) . " > " . $glyph->{'name'} . ":(" . join(" ", 1 .. ($#parts + 1)) . ") " . join(" ", ("_") x (@parts - 1)) . ";\n");
+        }
     }
     $fh->print("endpass;\nendtable;\n");
 }
@@ -356,3 +421,26 @@ EOT
     }
     $fh->print("endpass;\nendtable;\n");
 }
+
+sub has
+{
+    my ($cls, $val) = @_;
+    foreach (@{$cls})
+    { return 1 if ($_ == $val); }
+    return 0;
+}
+
+1;
+
+=head1 AUTHOR
+
+Martin Hosken L<Martin_Hosken@sil.org>. 
+
+=head1 LICENSING
+
+Copyright (c) 1998-2013, SIL International (http://www.sil.org)
+
+This module is released under the terms of the Artistic License 2.0.
+For details, see the full text of the license in the file LICENSE.
+
+=cut
